"""
This module is used to check if a user's code is equal to solution code and
note when the two differ.
"""

# parsing
import ast

# formatting
from .formatters import formatted
# checking functions
from .check_functions import standardize_arguments
# feedback
from .message_generators import (
    missing, not_expected, wrong_value, repeated_argument
)

# misc
from itertools import zip_longest
from typing import Dict, Any

def new_parent_source(tree: Any, last_parent: str) -> str:
    """Given an ast, the current tree, and the last parent source,
    return a new last parent source for facilitating feedback on problematic
    node.

    Parameters
    ----------
    tree : Any
        the ast.AST or a Python type we want to extract the source text for
    last_parent : str
        the nearest parent that can be converted to source text

    Returns
    -------
    str
        the new last_parent or the same
    """
    # only set the last_parent if:
    # 1) it is a somewhat complex AST which is defined as having more than 1 field, 
    # wich means that for e.g. "1" is not a good candidate since we're trying to support
    # calling out higher level expressions
    # 2) we can get a source text
    if (len(tree._fields) > 1 and 
        (issubclass(tree.__class__, ast.Expr) or issubclass(tree.__class__, ast.Call))):
        # attempt to get a new parent source text for feedback
        try:
            last_parent = formatted(tree)
        except:
            pass
    return last_parent
    
def check_children(left: Any, 
                   right: Any, 
                   line_info: Dict[str, int],
                   left_source: str = "", 
                   right_source: str = "",
                   last_parent: str = ""):
    """Checks children of two asts by iterating their fields

    Parameters
    ----------
    left : ast.AST
        the student code
    right : ast.AST
        the solution code
    line_info :  Dict[str, int]]
        holds line information about a particular node
    left_source : str, optional
        the source text for user code, by default ""
    right_source : str, optional
        the source text for the solution code, by default ""
    last_parent : str, optional
        the nearest parent that can be converted to source text, by default ""
    """
    lf, rf = ast.iter_fields(left), ast.iter_fields(right)
    # iterate through the children of both ASTs
    for left_field, right_field in zip_longest(lf, rf, fillvalue=""):
        left_name, left_values = left_field
        right_name, right_values = right_field
        # attempt to get a new parent source text for feedback
        # check that the name of the AST nodes match
        compare_node(left_name, right_name, line_info, last_parent)
        # recurse on values
        compare_ast(left_values, right_values, line_info, left_source, right_source, last_parent)

def compare_ast(left: Any, 
                right: Any, 
                line_info: Dict[str, int],
                left_source: str = "", 
                right_source: str = "",
                last_parent: str = "") -> None:
    """Compare two abstract syntax trees. Raise AssertionError as soon as they differ.

    Parameters
    ----------
    left : Any
        the student code
    right : Any
        the solution code
    line_info :  Dict[str, int]]
        holds line information about a particular node
    left_source : str
        the source text for user code, by default ""
    right_source : str
        the source text for the solution code, by default ""
    last_parent : str
        the nearest parent that can be converted to source text, by default ""
    """
    # to hold line information like line number, and original source
    line_info = {} if line_info is None else line_info
    # check types first
    compare_node_type(left, right, line_info)

    # the check AST, list of Expr, or a core data type
    if isinstance(left, ast.AST):
        # store line number for feedback
        line_info["left"] = getattr(
            left, "lineno", line_info.get("left", 1)
        )
        line_info["right"] = getattr(
            right, "lineno", line_info.get("right", 1)
        )
        # attempt to get a new parent source text for feedback
        last_parent = new_parent_source(left, last_parent)
        # for ast.Call we will raise error when there is either a problem 
        # with the function or the arguments do not match after standardization
        if isinstance(left, ast.Call):
            # check that the standardized left and right Calls are the same
            check_functions(left, right, line_info, left_source, right_source, last_parent)
        else:
            check_children(left, right, line_info, left_source, right_source, last_parent)
        
    elif isinstance(left, list):
        for left_child, right_child in zip_longest(left, right, fillvalue=""):
            # recurse on [Expr, ...]
            # attempt to get a new parent source text for feedback
            compare_ast(left_child, right_child, line_info, left_source, right_source, last_parent)
    else:
        compare_node(left, right, line_info, last_parent)

def check_functions(left_call: ast.Call, 
                   right_call: ast.Call, 
                   line_info: Dict[str, int],
                   left_source: str = "", 
                   right_source: str = "",
                   last_parent: str = ""):
    """Standardizes and compares two ast.Calls. Raise AssertionError as soon as they differ.

    Parameters
    ----------
    left_call : ast.Call
        the user function call
    right_call : ast.Call
        the solution function call
    line_info :  Dict[str, int]]
        holds line information about a particular node
    last_parent : str
        the nearest parent that can be converted to source text, by default ""
    left_source : str
        the source text for user code, by default ""
    right_source : str
        the source text for the solution code, by default ""
    """
    # TODO standardize_arguments args naming is confusing, perhaps refactor it.
    # standardize the left and right tree
    ls = standardize_arguments(left_call=left_call, right_call=right_call, left_source=left_source, right_source=right_source)
    rs = standardize_arguments(left_call=right_call, right_call=right_call, left_source=right_source, right_source=right_source)
    # if we don't have any arguments simply compare the two nodes
    if len(ls.keywords) == 0:
        check_children(left_call, right_call, line_info, left_source, right_source, last_parent)
    # else, check all of the arguments which are all in keywords after running
    # `standardize_arguments` to simplify checking
    if ls != None and rs != None:
        for l, r in zip_longest(ls.keywords, rs.keywords, fillvalue=""):
            # check if the student and solution keyword matches in their value
            wrong_value(l.value, r.value, line_info, formatted(l.value) == formatted(r.value), last_parent)
            # last parent node would be the value of the parameter if it's an expression/call
            last_parent = new_parent_source(l.value, last_parent)
    else:
        raise AssertionError("Foo-y! Something went wrong with function call checking.")
    
def compare_node_type(
        left: Any, 
        right: Any,
        line_info: Dict[str, int],
        last_parent: str = ""
    ) -> None:
    """Compare two ASTs' types and raise an exception with the line number if different.

    Parameters
    ----------
    left : Any
        an ast.Node, a custom type, or a python data type
    right : Any
        an ast.Node, a custom type, or a python data type
    line_info : Dict[str, int]
        holds line information about a particular node
    """
    # handle missing code
    if (isinstance(left, str) and left == '') and not isinstance(right, str):
        missing(left, right, line_info)
    elif left != '' and right == '':
        # handle case for an unexpected value
        not_expected(left, right, line_info)
    else:
        # or, wrong types
        wrong_value(left, right, line_info, type(left) == type(right), last_parent)

def compare_node(left: Any, right: Any, line_info: Dict[str, int], last_parent: str = ""):
    """Compare two objects of the same type and raise an exception with feedback if nodes differ.

        Parameters
        ----------
        left : Any
            an ast.Node, a custom type, or a python data type
        right : Any
            an ast.Node, a custom type, or a python data type
        line_info : Dict[str, int]
            holds line information about a particular node
        last_parent : str
            the nearest parent that can be converted to source text
    """
    wrong_value(left, right, line_info, left == right, last_parent)

def grade_code(student_code: str, solution_code: str):
    """Checks user and solution code and prints a message if they differ

    Parameters
    ----------
    student_code : str
        the user source code
    solution_code : str
        the solution source code

    Returns
    -------
    None
        If there are no issues
    str
        If user code differs from solution
    """
    try:
        # source node back
        if student_code == '':
            raise Exception("I didn't receive the student code.")
        elif solution_code == '':
            raise Exception("I didn't receive the solution code.")
        student = ast.parse(student_code)
        solution = ast.parse(solution_code)
        last_parent = [a for a in ast.walk(ast.parse(student))][1]
        compare_ast(
            student, 
            solution, 
            {}, 
            left_source = student_code, 
            right_source = solution_code, 
            last_parent = new_parent_source(last_parent, "")
        )
    except SyntaxError as e:
        message = str(e)
        # TODO figure out why we're getting unknown for diagnosis of argument
        if "repeated" in message:
            return repeated_argument(e)
    except AssertionError as e:
        return str(e) # back to either the python_grader or pygradethis_exercise_checker
    except Exception as e:
        return "There was a problem checking user or solution code: {}".format(e)



"""
This module contains functions to faciliate checking Python code or output.
"""
from typing import Any

from .pygradethis_exercise_checker import pygradethis_exercise_checker
from .conditions import Graded

def grade(
  *check_code: Graded,
  user_code: str = None, 
  solution_code: str = None,
  last_value: Any = None
  ) -> dict:
  """Python standalone verson of `pygradethis_exercise_checker`. This function does
  two things:
  - Do static code grading (AST) if both user and solution code is provided
    before the result grading.
  - Do result grading based on code output and Graded(s)

  Note: Currently, this is just a wrapper around `pygradethis_exercise_checker` and in the
  future we can make changes to handle standalone-specific details.

  Parameters
  ----------
  check_code: *Graded
      a variable number of Graded objects
  user_code : str, optional
      user source code, by default None
  solution_code : str, optional
      solution source code, by default None

  Returns
  -------
  dict
      a feedback dict:
      {
        message: str,
        correct=True|False,
        type="auto|success|info|warning|error|custom",
        location="append|prepend|replace"
      }
  """
  return pygradethis_exercise_checker(
    user_code = user_code,
    solution_code = solution_code,
    check_code = check_code,
    envir_prep = {},
    envir_result = {}
  )





"""
Module contains a generic AST node formatter and allows for custom formatters
for whichever data type you need.

To define a node formatter simply define another `_` function, register its 
type for the generic `formatted` function, and define the string to return:

```
@formatted.register(<your type>)
def _(node: ast.AST) -> str:
    <process your node>
    return "<your string>"
```

Alternatively, you can also call the `format_node` function to register your
function using one of two ways:

1. `format_node(int, <"a string">)`

2. `format_node(int, <a function or lambda that formats and returns string>)`

The basic data types and ast types have been supported but can be changed in
the respective register function definition, or overriden with `format_node`.
"""

import ast
from functools import singledispatch

from typing import Any

@singledispatch
def formatted(node: Any) -> str:
    """Generic function to return a formatted representation of an AST node"""
    try:
        return ast.unparse(node)
    except AttributeError:
        return "{}".format(type(node).__name__)

def format_node(node_type, string_format):
    """Registers a format function for a give node type
    
    The node_type must be a valid data type in Python.
    The string_format can be either a str or a function / lambda.
    """
    @formatted.register(node_type)
    def _(node: ast.AST) -> str:
        if callable(string_format):
            return string_format(node)
        else:
            return string_format

# binary operators
format_node(ast.Add, "+")
format_node(ast.Sub, "-")
format_node(ast.Mult, "*")
format_node(ast.MatMult, "@")
format_node(ast.Div, "/")
format_node(ast.FloorDiv, "//")
format_node(ast.Mod, "%")
format_node(ast.Pow, "**")
format_node(ast.LShift, "<<")
format_node(ast.RShift, ">>")
format_node(ast.BitOr, "|")
format_node(ast.BitXor, "^")
format_node(ast.BitAnd, "&")
# unary operators
format_node(ast.UAdd, "+")
format_node(ast.USub, "-")
format_node(ast.Invert, "~")
# bool operators
format_node(ast.And, "and")
format_node(ast.Or, "or")
format_node(ast.Not, "not")
# compare operators
format_node(ast.Eq, "==")
format_node(ast.NotEq, "!=")
format_node(ast.Lt, "<")
format_node(ast.LtE, "<=")
format_node(ast.Gt, ">")
format_node(ast.GtE, ">=")
format_node(ast.Is, "is")
format_node(ast.IsNot, "is not")
format_node(ast.In, "in")
format_node(ast.NotIn, "not in")

# core types ------------------------------------------

format_node(int, lambda x: str(x))
format_node(str, lambda x: "\"{}\"".format(x))
format_node(ast.Str, lambda x: "\"{}\"".format(x.s))
format_node(ast.Constant, lambda x: formatted(x.value))

@formatted.register(ast.List)
def _(node: ast.List) -> str:
    return "[" + ", ".join(formatted(e.value) for e in node.elts) + "]"

@formatted.register(ast.Lambda)
@formatted.register(ast.Compare)
@formatted.register(ast.BoolOp)
@formatted.register(ast.BinOp)
def _(node: ast.AST) -> str:
    return ast.unparse(node)

@formatted.register(list)
def _(node: list) -> str:
    return "{}".format(", ".join(formatted(e) for e in node))

@formatted.register(ast.keyword)
def _(node: ast.keyword) -> str:
    return "{}={}".format(node.arg, formatted(node.value))

@formatted.register(ast.Call)
def _(node: ast.Call) -> str:
    return "{}({})".format(
        formatted(node.func), 
        ", ".join(filter(None, [formatted(node.args), formatted(node.keywords)]))
    )

@formatted.register(ast.UnaryOp)
def _(node: ast.UnaryOp) -> str:
    return "{}{}".format(formatted(node.op), formatted(node.operand))

@formatted.register(ast.Module)
def _(node: ast.Module) -> str:
    """Formatting function for:
    
    ast.Module

    Parameters
    ----------
    node : ast.Module
        the Module node

    Returns
    -------
    str
        unparsed node
    """
    for b in node.body: 
        if hasattr(b, 'value'):
            return formatted(b.value)
        else:
            return formatted(b)
    return ""



import itertools
from copy import deepcopy
from dataclasses import dataclass
from typing import Optional, AnyStr

from lxml.etree import _Element as Element

from .find_utils import uses, flatten_list
from .grade_code_found import GradeCodeFound, get_node_source
from .ast_to_xml import xml
from .xml_utils import (
  compare_xml_nodes, literal, expr_xml_node
)

@dataclass
class Arg:
  """A dataclass to represent a function's positional argument"""
  name: Optional[str] = ''
  code: AnyStr = None
  xml_node: AnyStr = None

  def __str__(self) -> str:
    if self.name != '':
      return f"positional argument: {self.name} = {self.code}"
    else:
      return f"positional argument: {self.code}"

@dataclass
class KWArg:
  """A dataclass to represent a function's keyword argument"""
  name: Optional[str] = ''
  code: AnyStr = None
  value_xml_node: AnyStr = None
  # this field is so we can refer to originating keyword XML Element
  # when it comes to match and store keyword arguments for `find_arguments()`
  keyword_xml_node: AnyStr = None
  
  def __str__(self) -> str:
    return f"keyword argument: {self.name} = {self.code}"

@dataclass
class ArgList:
  """A dataclass to represent a function's argument or keyword argument"""
  args: list[Arg]

  def __init__(self, args: list[Arg] = None):
    self.args = [] if args is None else args

def args(*args: AnyStr, **kwargs: AnyStr) -> ArgList:
  """Returns a list of Arg objects for positional (unnamed) argument values 
  or keyword arguments to facilitate finding arguments in function calls.

  Returns
  -------
  ArgList
      an ArgList object that holds a list of Arg(s)
  """
  all_args = []

  # first add the positional arguments
  if len(args) > 0:
    positional = list(args)
    all_args.extend([
      # because Python eagerly evaluates all arguments passed in to args()
      # we have to specially handle the quoting of them: 
      # - if the argument was meant to be a `str` we double quote `"` it
      # - anything else we just wrap into a normal string
      Arg(
        code = arg.source if isinstance(arg, literal) else arg,
        xml_node = expr_xml_node(arg)
      )
      for arg in positional
    ])

  # then, add the keyword arguments
  if len(kwargs) > 0:
    all_kwargs = list(kwargs.items())
    # this is extracting the name = value
    all_args.extend(
      # similar logic to positional arguments here, where kwargs value strings
      # are quoted if it's meant to be a `str`
      KWArg(
        name = kwarg[0],
        code = kwarg[1].source if isinstance(kwarg[1], literal) else kwarg[1],
        value_xml_node = expr_xml_node(kwarg[1])
      )
      for kwarg in all_kwargs
    )

  return ArgList(all_args)

def make_kwargs(code: str, kwarg_xml_nodes: list[Element]) -> list[KWArg]:
  """Return KWArg(s) based on <keyword> XML Element(s)

  Parameters
  ----------
  code : str
      the source code text
  kwarg_xml_nodes : list[Element]
      list of XML Elements representing keyword

  Returns
  -------
  list[KWArg]
      a list of KWArgs representing those keywords
  """
  code_kwargs_list = []
  for kwarg_node in kwarg_xml_nodes:
    # keyword argument name
    kwarg_names = [
      x.text for x in kwarg_node.xpath("./arg")
    ]

    # get the textual form for values
    kwarg_values = [v for v in kwarg_node.xpath("./value/*")]
    kwargs_values_strings = [
      get_node_source(code, kw)
      for kw in kwarg_values
    ]

    # combine them into pairs
    kwargs_pairs = list(zip(kwarg_names, kwargs_values_strings))

    # construct the list of Arg(s) from the keyword pairs
    code_kwargs_list.append([
      KWArg(
        name = k,
        code = v,
        value_xml_node = expr_xml_node(v),
        keyword_xml_node = kwarg_node
      )
      for k, v in kwargs_pairs
    ])
  return flatten_list(code_kwargs_list)

def find_arguments(
  code: str | GradeCodeFound,
  match: ArgList = ArgList(),
  recurse: bool = True
) -> GradeCodeFound:
  """Find arguments of function calls in the code.

  Parameters
  ----------
  code : str | GradeCodeFound
      a string for the source code, or a GradeCodeFound for chaining queries
  match : ArgList
      an ArgList representing the arguments to match against
  recurse : bool
      a flag that requests whether or not to look for arguments
      within function call arguments for a previous 'function' request, by default False

  Returns
  -------
  GradeCodeFound
      a GradeCoundFound object that holds the list of previous results
      and the current query results if any, by default True
  
  Examples
  --------
  >>> code = "sum([1, round(2.5), 3])\nprint('Hello', 'World!', sep=', ')"
  >>> find_arguments(code)
  ── pygradecode found ──
  print(1, "2", sep=", ", end=foo(x = ["\n"]))

  ── Request ──
  arguments 
  Found 5 results.

  ── Result 1 ──
  1

  ── Result 2 ──
  "2"

  ── Result 3 ──
  sep=", "

  ── Result 4 ──
  end=foo(x = ["\n"]))

  ── Result 5 ──
  x = ["\n"])
  """
  if not isinstance(code, str) and not isinstance(code, GradeCodeFound):
    raise Exception("`code` should be a `str` or `GradeCodeFound`")
  
  gcf = deepcopy(code) if isinstance(code, GradeCodeFound) else GradeCodeFound(code)
  xml_tree = xml(gcf.source) if isinstance(code, GradeCodeFound) else xml(code)

  request_type = 'arguments'
  # TODO create a __str__ representation for ArgList
  request = ''
  results = []

  # Note: we have to encode the code to escape it properly
  match_args = match.args

  # get previous function calls in results if any, otherwise get all function calls
  all_call_xml_nodes = []
  if gcf.has_previous_request():
    for node in gcf.last_result:
      if recurse:
        all_call_xml_nodes.append(node.xpath('//Call'))
      else:
        all_call_xml_nodes.append(node)
    all_call_xml_nodes = list(itertools.chain(*all_call_xml_nodes))
  else:
    all_call_xml_nodes = xml_tree.xpath("//Call")

  # if there are no arguments to look for in particular, look for
  # all of the arguments in code
  if len(match_args) == 0:
    results = [
      node.xpath("./args/*|./keywords/*")
      for node in all_call_xml_nodes
    ]
    results = list(itertools.chain(*results))
  else:
    # Extracting each keyword pair is a bit tricky because a value
    # could be any ast node so there is no way to have one query
    # satisfy all types of nodes. So, instead we look at the
    # <keyword> and its <value> node source text to reconstruct them.
    # TODO if a gcf was passed when there was a previous request
    # we need to iterate through gcf.results to repeat this process
    # so we can make all of the Arg/KWArg

    # extract the positional args
    all_arg_xml_nodes = flatten_list([
      call.xpath("./args/*")
      for call in all_call_xml_nodes
    ])
    # extract the kwargs args
    all_kwarg_xml_nodes = flatten_list([
      call.xpath("./keywords/*")
      for call in all_call_xml_nodes
    ])

    # construct Arg() for positional arguments
    code_args_list = [
      Arg(
        code = get_node_source(code, arg),
        xml_node = deepcopy(arg)
      )
      for arg in all_arg_xml_nodes
    ]

    # extract the keyword args
    code_kwargs_list = make_kwargs(code, all_kwarg_xml_nodes)
    # combine the args with the kwargs
    all_code_args = code_args_list + code_kwargs_list
      
    # check that each match Arg is in code's [Arg]
    results = get_matched_arguments(match_args, all_code_args)

  # TODO make the chaining work well
  return gcf.push(request_type=request_type, request=request, result=results)

def get_matched_arguments(
  match_args: list[Arg | KWArg],
  all_code_args: list[Arg | KWArg]
 ) -> list[Element]:
  """Return all code XML elements that match given the list of match and code arguments.

  Parameters
  ----------
  match_args : list[Arg  |  KWArg]
      a list with Arg and KWArgs representing positional and keyword arguments
      to match against
  all_code_args : list[Arg  |  KWArg]
      a list with Arg and KWArgs representing positional and keyword arguments
      of the code

  Returns
  -------
  list[Element]
      a list of XML Element(s)
  """
  results = []
  for match_arg in match_args:
    # for argument, element in args_spec:
    for argument in all_code_args:
      # if match Arg Element is the same as the code Arg Element
      # add the code Arg Element as a result
      if type(match_arg) != type(argument):
        continue
      elif isinstance(match_arg, KWArg) and isinstance(argument, KWArg):
        # if we're comparing keyword arguments 
        # check value nodes match
        value_nodes_match = compare_xml_nodes(
          match_arg.value_xml_node,
          argument.value_xml_node
        )
        # check the kwarg name too
        arg_names_match = match_arg.name == argument.name
        if arg_names_match and value_nodes_match:
          results.append(argument.value_xml_node)
      elif compare_xml_nodes(match_arg.xml_node, argument.xml_node):
        results.append(argument.xml_node)
  return results

def uses_argument(code: str, match: str = "") -> bool:
  """Check if the code has arguments.

  Parameters
  ----------
  code : str
      the source code
  match : str, optional
      argument name(s), by default None

  Returns
  -------
  bool
      True if found, False otherwise

  Examples
  --------
  >>> code = "print('Hello', 'World!', sep=', ')"
  >>> uses_argument(code, "sum")
  True
  >>> uses_argument(code, "round")
  False
  """
  return uses(find_arguments, code, match)



# core
import sys
from io import StringIO
from itertools import zip_longest
from collections import namedtuple
from typing import Any, Tuple

from .conditions import Graded
# testing
import unittest
try:
    # attempt to import `pandas` related asserts
    from pandas.testing import assert_frame_equal, assert_series_equal
except:
    pass

class TestCondition(unittest.TestCase):
    """Custom TestCase to do asserts on Graded(s) and incorporate
    custom asserts from libraries like pandas for dataframe checking.
    """
    def __init__(self, test_name: str, condition: Graded):
        super(TestCondition, self).__init__(test_name)
        self.condition = condition
        
    ## Test Case for `Graded`
    def test_condition(self):
        # Note: currently the `correct` field from TestCase tuple isn't used
        # and is only included to avoid a tuple unpacking error
        ID, actual, expected, correct = self.condition
        try:
            if (actual.__class__.__name__ == "DataFrame" and 
                expected.__class__.__name__ == "DataFrame" and
                "pandas" in sys.modules):
                assert_frame_equal(actual, expected)
            else:
                # skip if it's a plain pass or fail condition
                if expected is None:
                    pass
                else:
                    self.assertEqual(actual, expected)
        except Exception:
            # Note: this ID isn't really being used at the moment for error
            # checking
            self.fail(ID)

# we want custom result for `TextTestRunner`
class ConditionTestResult(unittest.TextTestResult):
    """Custom test result where we maintained a matched test ID list to reference
    later when determining matched conditions.
    """
    def __init__(self, stream, descriptions: bool, verbosity: int):
        super(ConditionTestResult, self).__init__(stream, descriptions, verbosity)
        self.matched = []
        self.num_correct = 0

    def addSuccess(self, test: TestCondition):
        super(ConditionTestResult, self).addSuccess(test)
        # add matched case
        self.matched.append(test.condition.id)
        if test.condition.correct:
            self.num_correct += 1

def test_conditions(*conditions: Graded, 
                    user_result: Any = None) -> Tuple[bool, Graded]:
    # create a test suite
    suite = unittest.TestSuite()
    # convenience tuple for structuring test cases
    TestCase = namedtuple("TestCase", ["id", "actual", "expected", "correct"])
    # make TestCondition based on GradeCondition(s)
    for i, c in enumerate(conditions):
        suite.addTest(
            TestCondition(
                'test_condition', TestCase(i, user_result, c['x'], c['correct'])
            )
        )
    # capture current sys.stderr buffer (default one)
    default_buffer = sys.stderr
    # capture test print output for AssertionError(s)
    sys.stderr = StringIO()
    # run suite
    results = unittest.TextTestRunner(resultclass = ConditionTestResult).run(suite)
    # restore stderr to buffer so we catch any errors on our matching code below
    sys.stderr = default_buffer
    
    # get matched conditions by subsetting conditions list with matched index list
    matched_conditions = list(map(lambda i: conditions[i], results.matched))
    incorrect_match = None
    if len(matched_conditions) > 0:
        # return on the first `pass_if_equals`
        for mc in matched_conditions:
            if mc['correct']:
                return mc
            elif incorrect_match == None:
                # keep record of the first `fail_if_equals` to return (if any)
                incorrect_match = mc
    else: # no match
        # for now just return failing
        # we will need to revisit the grading flow later to handle cases where there are no
        # matching grading conditions
        return Graded(
                x = None,
                message = "",
                correct = False,
                type = "error"
            )
    return incorrect_match



"""Module to find attributes, method calls, properties in Python code."""

from copy import deepcopy

from lxml.etree import _Element as Element

from .ast_to_xml import xml
from .grade_code_found import GradeCodeFound
from .find_utils import uses, flatten_list

def find_attributes(code: str | GradeCodeFound) -> GradeCodeFound:
  """Find top-level attribute access in the code.

  Attributes could be a property, a method call, or any other AST
  types where user calls <object>.<attribute>

  Unlike some other find_*() functions, this one simply grabs all
  instances without matching on anything in particular. The reason for
  this is because it is hard to specify matches for any arbitrary ASTs.
  To target specific things, use other find functions following this
  to narrow down the query.

  Parameters
  ----------
  code : str | GradeCodeFound
      a string for the source code, or a GradeCodeFound for chaining queries

  Returns
  -------
  GradeCodeFound
      a GradeCoundFound object that holds the list of previous results
      and the current query results if any
  """
  if not isinstance(code, str) and not isinstance(code, GradeCodeFound):
    raise Exception("`code` should be a `str` or `GradeCodeFound`")
  
  gcf = deepcopy(code) if isinstance(code, GradeCodeFound) else GradeCodeFound(code)
  xml_tree = xml(gcf.source) if isinstance(code, GradeCodeFound) else xml(code)

  request_type = 'attributes'
  request = ''
  result = []

  xpath_query = ".//Attribute"

  attrs = []
  if gcf.has_previous_request():
    for node in gcf.last_result:  
      attrs.extend(node.xpath(xpath_query))
    result.extend(attrs)
  else:
    attrs = xml_tree.xpath(xpath_query)
    result.extend(
      flatten_list(
        a.xpath("../..") # go up to the grandparent node (to go above <value> or <func>)
        for a in attrs
      )
    )

  return gcf.push(request_type=request_type, request=request, result=result)

def uses_attributes(code: str) -> bool:
  """Check if the code uses attributes.

  Parameters
  ----------
  code : str
      the source code

  Returns
  -------
  bool
      True if found, False otherwise

  Examples
  --------
  >>> code = 'df.shape'
  >>> uses_attributes(code)
  True
  >>> code = 'df.head()'
  >>> uses_attributes(code)
  True
  >>> code = 'df.loc[:,'foo']'
  >>> uses_attributes(code)
  True
  >>> code = 'sum([1, 2, 3])'
  >>> uses_properties(code)
  False
  """
  return uses(find_attributes, code)

def find_properties(code: str | GradeCodeFound, match: str = "") -> GradeCodeFound:
  """Find property access in the code.

  For example, one could access a field like `df.shape`

  Parameters
  ----------
  code : str | GradeCodeFound
      a string for the source code, or a GradeCodeFound for chaining queries
  match : str, optional
      a particular property, by default None

  Returns
  -------
  GradeCodeFound
      a GradeCoundFound object that holds the list of previous results
      and the current query results if any
  """
  if not isinstance(code, str) and not isinstance(code, GradeCodeFound):
    raise Exception("`code` should be a `str` or `GradeCodeFound`")
  
  gcf = deepcopy(code) if isinstance(code, GradeCodeFound) else GradeCodeFound(code)
  xml_tree = xml(gcf.source) if isinstance(code, GradeCodeFound) else xml(code)

  request_type = 'property'
  request = match
  result = []

  if match == "":
    xpath_query = ".//Expr/*/Attribute/../.." 
    # TODO we have duplication of this logic to handle previous requests if it exists
    # let's refactor it out into a function if possible
    if gcf.has_previous_request():
      for node in gcf.last_result:  
        result.extend(flatten_list(node.xpath(xpath_query)))
    else:
      result.extend(xml_tree.xpath(xpath_query))
  else:
    # TODO support a list of matches
    xpath_query = f".//attr[text()='{match}']/ancestor::Expr"
    if gcf.has_previous_request():
      for node in gcf.last_result:  
        result.extend(flatten_list(node.xpath(xpath_query)))
    else:
      result.extend(xml_tree.xpath(xpath_query))

  return gcf.push(request_type=request_type, request=request, result=result)

def uses_properties(code: str, match: str = "") -> bool:
  """Check if the code uses properties.

  Parameters
  ----------
  code : str
      the source code
  match : str, optional
      the property name(s), by default None

  Returns
  -------
  bool
      True if found, False otherwise

  Examples
  --------
  >>> code = 'df.shape'
  >>> uses_properties(code)
  True
  >>> code = 'df.head()'
  >>> uses_properties(code)
  False
  >>> code = 'df.empty'
  >>> uses_properties(code, 'empty')
  True
  >>> uses_properties(code, 'shape')
  False
  """
  return uses(find_properties, code, match)

def find_method_chains(code: str | GradeCodeFound) -> GradeCodeFound:
  """Find method chains in the code.

  Note: This function will search for the top-level chains, so it will not
  return chains that are inside other expressions.

  Parameters
  ----------
  code : str | GradeCodeFound
      a string for the source code, or a GradeCodeFound for chaining queries

  Returns
  -------
  GradeCodeFound
      a GradeCoundFound object that holds the list of previous results
      and the current query results if any
  """
  if not isinstance(code, str) and not isinstance(code, GradeCodeFound):
    raise Exception("`code` should be a `str` or `GradeCodeFound`")
  
  gcf = deepcopy(code) if isinstance(code, GradeCodeFound) else GradeCodeFound(code)
  xml_tree = xml(gcf.source) if isinstance(code, GradeCodeFound) else xml(code)

  request_type = 'property'
  request = ''

  result = [
    a
    # if an Attribute is followed by another Attribute, we have a chain
    for a in xml_tree.xpath(".//Attribute") if a.xpath("(.//Attribute)[1]")
  ]

  return gcf.push(request_type=request_type, request=request, result=result)

def uses_method_chains(code: str) -> bool:
  """Check if the code uses method chains.

  Parameters
  ----------
  code : str
      the source code
  match : str, optional
      function name(s), by default None

  Returns
  -------
  bool
      True if found, False otherwise

  Examples
  --------
  >>> code = 'df.head().tail()'
  >>> uses_function(code)
  True
  >>> code = 'df.head()'
  >>> uses_function(code)
  False
  """
  return uses(find_method_chains, code)



"""Functions that format and style various objects to pretty print."""

import builtins
from dataclasses import dataclass
from functools import singledispatch

from lxml.etree import _Element as Element
from rich.console import Console

# pgc_print() is a generic print() method
@singledispatch
def pgc_print(arg):
  builtins.print(arg)

@dataclass
class FormattedText:
  text: str

@pgc_print.register
def _(arg: FormattedText, console: Console = Console(color_system='standard')):
  console.print(arg.text)

# Override the default print() with my_print
print = pgc_print

def format_text(code: str, target_node: Element) -> FormattedText:

  location = target_node.attrib
  end_line = int(location['end_lineno']) - 1

  code_lines = code.splitlines()

  # get the target line
  # NOTE: we might have different heuristics for different types of nodes
  # but generally when a particular node spans multiple lines
  # it might suffice to just focus on the most pertinent lines instead
  # of highlight blocks of code
  target_line = code_lines[end_line]

  # get column offset
  col_start = int(location['col_offset'])
  col_end = int(location['end_col_offset'])

  # beginning string is all previous lines
  prev_str = "\n".join(code_lines[:end_line])
  
  beginning_line_str = target_line[:col_start]
  formatted_str = (
    "[u bold]" + 
    target_line[col_start] + 
    target_line[col_start + 1:col_end] + 
    "[/u bold]"
  )
  end_line_str = target_line[col_end:]

  # end string is the rest of the lines

  remaining_str = "\n".join(code_lines[end_line + 1:])

  return FormattedText(
    f"{prev_str}\n{beginning_line_str}{formatted_str}{end_line_str}\n{remaining_str}"
  )



"""
Module with useful functions.
"""

from typing import Tuple
import ast

class NotSet:
    """Sentinel value to distinguish a special type of None from literal
    None. This is used specifically for get_last_value function.
    """
    def __repr__(self):
        return "None"

NONE = NotSet()

def parse_code(input: str | list[str]) -> str:
    """Tries to parse code represented as string or list of strings

    Parameters
    ----------
    input : str | list[str]
        either a str or a list of str

    Returns
    -------
    str
        the formatted string

    Raises
    ------
    SyntaxError
        if there are any parsing issues
    """
    if input is None:
        return input
    try:
        simple = "".join(input)
        ast.parse(simple)
        return simple
    except SyntaxError as e:
        if "EOF" in str(e):
            return "\n".join(input)
        else:
            raise SyntaxError("Problem parsing your code!")

def get_last_value(script: str, envir: dict) -> Tuple:
    """Evaluate Python code and return the value of the last line of code.

    Parameters
    ----------
    script : str
        A string of Python code
    envir : dict
        A dictionary for the environment in which to evaluate code

    Returns
    -------
    Tuple
        The Python envir which contains all objects created in execution of envir,
        and a last value which is a Python object, or `NotSet` if the last line is not an expression.

    Examples
    --------
    >>> get_last_value("x = 2; x + 1")
    3

    >>> get_last_value("2 + 2; x = 2") # returns NONE (sentinel value for None)
    """
    stmts = list(ast.iter_child_nodes(ast.parse(script)))
    # iteratively walk through AST and execute statements and expressions while storing the last value
    last_value = None
    for s in stmts:
      if isinstance(s, ast.Expr):
        last_value = eval(compile(ast.Expression(body=s.value), filename="<ast>", mode="eval"), envir)
      else:
        exec(compile(ast.Module(body=[s], type_ignores=[]), filename="<ast>", mode="exec"), envir)
        last_value = NONE
    
    return envir, last_value

def get_envir_diff(x: dict, y: dict) -> list[str]:
  """Find the difference of variables created in prepped versus student environment

  This is used to find new variables introduced in `.py_envir_result` or `.py_envir_solution`
  excluding the `.py_envir_prep` setup environment

  Parameters
  ----------
  x : dict
      the `.py_envir_prep` module
  y : dict
      the `.py_envir_result` or `.py_envir_solution` module

  Returns
  -------
  list[str]
      a list of the variable names created in result environment
  """
  return list(set(y.keys()) - set(x.keys()))



from collections import namedtuple
from typing import Optional
from lxml.etree import _Element as Element
from rich.console import Console

from .ast_to_xml import get_source_lines
from .find_utils import get_ancestor_node
from .highlight_text import format_text, pgc_print

# `type` hold the types of requests (e.g. 'function' for find_functions())
# `request` hold the specific requests (e.g. 'sum' for find_functions())
# `result` holds the list of XML element results for each request
QueryResult = namedtuple("QueryResult", ['type', 'request', 'result'])

class GradeCodeFound:
  source: str
  results: list[QueryResult]

  def __init__(self, code: str = "", results: Optional[QueryResult] = None) -> None:
    self.source = code
    self.results = results if results is not None else []

  def push(
    self, 
    request_type: str, 
    request: str, 
    result: list[Element]
  ) -> 'GradeCodeFound':
    self.results.append(
      QueryResult(type=request_type, request=request, result=result)
    )
    return self
  
  def has_previous_request(self) -> bool:
    return len(self.results) > 0
  
  @property
  def last_result(self) -> list[Element]:
    if self.has_previous_request():
      return self.results[-1].result
    else:
      return []
  
  def get_last_state(self) -> QueryResult:
    if self.has_previous_request():
      return self.results[-1]
    return None

  def __repr__(self):
    pgc_print(self)
    return ""

@pgc_print.register
def _(arg: GradeCodeFound, console: Console = Console(color_system='standard')):
  last_state = arg.get_last_state()

  if last_state is None:
    return

  last_type, last_request, last_result = last_state

  # code
  console.print(f"── pygradecode found ──\n{arg.source.strip()}\n")

  # type of request and specific request (if any)
  console.print(f"── Request ──\n{last_type} {last_request}")

  console.print(
    f"Found {len(last_result)} {'results' if len(last_result) > 1 else 'result'}.\n"
  )

  # result
  for i, element in enumerate(last_result):
    console.print(f"── Result {i + 1} ──")
    formatted = format_text(arg.source, element)
    console.print(formatted.text, end="\n\n")

def get_node_source(code: str | GradeCodeFound, node: Optional[Element]) -> str:
  if node is None:
    return ''

  if isinstance(code, GradeCodeFound):
    code = code.source

  target_code = get_source_lines(
    src_lines = code.splitlines(),
    node = node,
    dedent = False # we want to preserve the whitespace
  )

  try:
    node_with_location = get_ancestor_node(node)
    start_col = int(node_with_location.attrib['col_offset'])
    end_col = int(node_with_location.attrib['end_col_offset'])
    return target_code[start_col:end_col].encode('raw_unicode_escape').decode()
  except (Exception, ValueError):
    pass


"""
This module is used to check python code output.
"""

from typing import Any, List

from .conditions import Graded
from .result_checker import test_conditions

# TODO add the glue type messages to construct custom messages
# glue_correct = getOption("gradethis_glue_correct"),
# glue_incorrect = getOption("gradethis_glue_incorrect")
def grade_result(
    *conditions: List[Graded], 
    user_result: Any = None
  ) -> dict | str:
  """This function checks the user's code output against the list of conditions.

  Parameters
  ----------
  conditions: *Graded
      a variable number of Graded objects
  user_result : Any, optional
      user's code output, by default None

  Returns
  -------
  Union[dict, str]
      dict, if there are no issues
      str, an error message if Graded(s) is not supplied

  Raises
  ------
  Exception
      if GraderConditions are not passed in for conditions
  """
  all_conditions = [c for c in conditions if isinstance(c, Graded)]
  if conditions == None or len(all_conditions) == 0:
    raise Exception(
      "At least one condition object (e.g., `pass_if_equals()`, "
      "`fail_if_equals()`, `python_condition()`) must be provided to"
      "`grade_result()`"
    )
  if user_result is not None:
    return test_conditions(*conditions, user_result = user_result)
  return conditions



from .ast_to_xml import xml
from .find_utils import uses
from .grade_code_found import GradeCodeFound

# operator symbol to ast node name
OPERATORS = {
  # unary ops
  '~': 'Invert', 'not': 'Not',
  # infix operators
  'and': 'And', 'or': 'Or',
  '+': 'Add',
  '-': 'Sub',
  '*': 'Mult',
  '@': 'MatMult',
  '/': 'Div',
  '%': 'Mod',
  '**': 'Pow',
  '<<': 'LShift',
  '>>': 'RShift',
  '|': 'BitOr',
  '^': 'BitXor',
  '&': 'BitAnd',
  '//': 'FloorDiv',
  # cmp ops
  '==': 'Eq', '!=': 'NotEq', '<': 'Lt', '<=': 'LtE', '>': 'Gt', '>=': 'GtE',
  'is': 'Is', 'is not': 'IsNot', 'in': 'In', 'not in': 'NotIn'
}

# unaryops
# NOTE: we separate these because +/- are unary operators that
# could get confused for the Add/Sub
UNARY_OPS = {'+': 'UAdd', '-': 'USub'}

def get_operator(op: str) -> list[str]:
  # check if op is valid
  if op not in OPERATORS and op not in UNARY_OPS:
    raise Exception(f"{op} is not a valid operator in Python!")

  # we may get an overlap of operators as a Binary or Unary operator
  # so we will try to combine the valid ones in a list
  return [o for o in [OPERATORS.get(op), UNARY_OPS.get(op)] if o is not None]

def find_operators(code: str | GradeCodeFound, match: str = "") -> GradeCodeFound:
  """Find operators in the code.

  NOTE: currently, the output is not super helpful because we're
  printing the first ancestor which has the location attributes
  to return source text. In the future, this will be solved by
  incorporating `rich` to bold/underline within the original code text.

  Parameters
  ----------
  code : str
      the source code
  match : str, optional
      a particular operator name, by default None

  Returns
  -------
  list[Element]
      list of XML elements corresponding to operators

  Examples
  --------
  >>> code = "-1 + 2 * 3 // 4"
  >>> find_operators(code)
  ── pygradecode found ──
  -1 + 2 * 3 // 4

  ── Request ──
  operators 
  Found 4 results.

  ── Result 1 ──
  -1

  ── Result 2 ──
  -1 + 2 * 3 // 4

  ── Result 3 ──
  2 * 3

  ── Result 4 ──
  2 * 3 // 4
  >>> find_operators(code, "-")
  ── pygradecode found ──
  -1 + 2 * 3 // 4

  ── Request ──
  operators -
  Found 1 result.

  ── Result 1 ──
  -1
  >>> find_operators(code, "//")
  ── pygradecode found ──
  -1 + 2 * 3 // 4

  ── Request ──
  operators //
  Found 1 result.

  ── Result 1 ──
  2 * 3 // 4
  """
  if not isinstance(code, str) and not isinstance(code, GradeCodeFound):
    raise Exception("`code` should be a `str` or `GradeCodeFound`")
  
  gcf = code if isinstance(code, GradeCodeFound) else GradeCodeFound(code)
  xml_tree = xml(gcf.source) if isinstance(code, GradeCodeFound) else xml(code)

  request_type = 'operators'
  request = match
  result = []
  
  if match != "":
    matched_ops = get_operator(match)
    # construct query to grab the specific matched operators
    xpath = "|".join(f"//op/{mo}" for mo in matched_ops)
    result = xml_tree.xpath(xpath)
  else:
    result = xml_tree.xpath("//op/*")

  return gcf.push(request_type=request_type, request=request, result=result)



"""
This is the `learnr` version of `python_grader` module and contains
the `pygradethis_exercise_checker` function called by the `gradethispython` R wrapper
package
"""
from copy import copy
from typing import Any, Union, List, Tuple

from .grade_result import grade_result
from .grade_code import grade_code
from .conditions import *
from .feedback import praise, encourage
from .utils import parse_code

def graded(graded: Union[str, dict]):
  if (graded is not None):
    correct = graded['correct']
    # correct
    if correct:
      return dict(
        message = "{} {}".format(praise(), graded['message']),
        correct = True,
        type = "success", 
        location = "append"
      )
    # incorrect
    return dict(
      message = "{} {}".format(encourage(), graded['message']),
      correct = False,
      type = "error", 
      location = "append"
    )
  # if there was none of the conditions matched, return error by default
  elif not graded:
    return dict(
      message = "{}".format(encourage()), 
      correct = False, 
      type = "error",
      location = "append"
    )

def pygradethis_exercise_checker(label: str = None,
                        solution_code: str = None,
                        user_code: str = None,
                        check_code: List[str] = None,
                        envir_result: dict = {},
                        evaluate_result: List[str] = [],
                        envir_prep: dict = {},
                        last_value: Any = None) -> dict:
  """This function mirrors the `grade_learnr` function from {gradethis} package so that
  we can check Python exercises. This function does two things:
  - Do static code grading (AST) if both user and solution code is provided
    before the result grading.
  - Do result grading based on code output and Graded(s)

  Parameters
  ----------
  label : str, optional
      exercise label, by default None
  solution_code : str, optional
      solution source code, by default None
  user_code : str, optional
      user source code, by default None
  check_code : List[str], optional
      checking code, by default None
  envir_result : dict, optional
      environment containing execution results, by default None
  evaluate_result : List[str], optional
      evaluation results, by default None
  envir_prep : dict, optional
      environment associated with exercise, by default None
  last_value : Any, optional
      the last value of exercise, by default None

  Returns
  -------
  dict
      a feedback dict:
      {
        message = str,
        correct = True|False,
        type = "auto|success|info|warning|error|custom",
        location = "append|prepend|replace"
      }
  """
  # check if there is user_code
  if user_code and "".join(user_code) == "":
    return dict(
      message="I didn't receive your code. Did you write any?",
      correct=False,
      type="error",
      location="append"
    )
  # if there is check code and solution code, check if there is a solution code
  if (check_code and solution_code) and "".join(solution_code) == "":
    return dict(
      message="No solution is provided for this exercise.",
      correct=True,
      type="info",
      location="append"
    )

  # TODO input scrubbing for malicious code
  # parse user, check and solution code

  # TODO validate checking code after the final check source is constructed
  # the final checking code includes the grading modules and stores the final grade in
  # a variable that we can reference later
  final_check_source = f"__result__ = {check_code}"
  
  # prep the dictionary that will hold imports, the variables passed into this function, and 
  # everything else that learnr stored into `envir_prep`
  # NOTE: in the future, we could abstract and have an Exercise class that would have envir bits 
  # and could have functions to do the checking + check flows
  if envir_prep is not None:
    graded_envir = dict(globals(), **locals(), **envir_prep)
  else:
    graded_envir = dict(globals(), **locals())
  graded_envir = {k: v for k, v in graded_envir.items() if v is not None}

  # evaluate exercise and check code output
  try:
    # NOTE: eventually this will have to follow the gradethis grading flow where check code
    # can either contain the grading the result or the code and use the student's result
    # evaluate check code and return the result and a Graded list structure
    exec(final_check_source, graded_envir)
    # extract the result out of the environment
    result = graded_envir['__result__']
  except Exception as e:
    # TODO somehow trickle up the specific error message?
    return dict(
      message="Error occured while checking the submission. {}".format(str(e)),
      correct=False,
      type="warning",
      location="append"
    )
  # return a dict for learnr to process for feedback
  return graded(result)



from copy import copy
from dataclasses import dataclass
from typing import Any

from lxml import etree as ET
from lxml.etree import _Element as Element
from lxml.doctestcompare import LXMLOutputChecker

from .ast_to_xml import xml

# low-level functions ----

@dataclass
class literal:
  source: str

  def __init__(self, source) -> None:
    if not isinstance(source, str):
      raise ValueError(f"`source` should be a `str`")
    self.source = source

def strip_node_location(n: Element) -> str:
  if n.attrib:
    n.set('lineno', '')
    n.set('end_lineno', '')
    n.set('col_offset', '')
    n.set('end_col_offset', '')
  return n

def xml_strip_location(node: Element) -> Element:
  """Strip the location information from an XML Element and all of its descendants.

  Parameters
  ----------
  node : Element
      the XML Element node

  Returns
  -------
  Element
      the same XML Element node with location information set to ""

  Raises
  ------
  Exception
      when a `None` is passed in for `node`
  """
  if node is None:
    raise Exception("cannot strip locations of a `None` Element")
  
  # strip the location information
  strip_node_location(node)
  
  for n in node.iterdescendants():
    strip_node_location(n)
    
  return node

def expr_xml_node(arg_code: str) -> Element:
  if (not isinstance(arg_code, str) and not isinstance(arg_code, literal)) or arg_code == '':
    raise Exception(
      f"`arg_code` for `py_args()` needs to be a string but was {type(arg_code)}"
    )

  if isinstance(arg_code, literal):
    arg_code = f"'{arg_code.source}'"
  else:
    arg_code = arg_code.replace(" ", "").encode('raw_unicode_escape').decode()

  try:
    xml_tree = xml(arg_code)
  except SyntaxError:
    raise Exception(
      f"The matching argument value for `py_args()` is not a valid Python expression string"
    )
  
  nodes = xml_tree.xpath('.//Expr/value/*[1]')
  
  return nodes.pop() if len(nodes) > 0 else None

def compare_xml_nodes(n1: Any, n2: Any):
  if n1 is None or n2 is None:
    return False
  
  # use the builtin LXMLOutputChecker
  # Note: we could in the future make an `xml_checker` param
  # to swap in any checker that has a `.compare_docs()` method
  xml_checker = LXMLOutputChecker()

  try:
    # strip locations on both and compare the two elements/docs
    n1 = copy(n1)
    n2 = copy(n2)

    result = xml_checker.compare_docs(
      xml_strip_location(n1),
      xml_strip_location(n2)
    )
    return result
  except Exception:
    pass
  
  return False

# XML Print methods ----

def prettify(xml_tree: Element) -> None:
  # Note: tostring() returns bytes, so we decode() it
  print(ET.tostring(xml_tree, pretty_print=True).decode())


from pygradethis import check_functions
from pygradethis import conditions
from pygradethis import feedback
from pygradethis import formatters
from pygradethis import grade_code
from pygradethis import grade_result
from pygradethis import message_generators
from pygradethis import pygradethis_exercise_checker
from pygradethis import python_grader
from pygradethis import result_checker
from pygradethis import utils

__version__ = "0.4.0"


"""
This module contains functions to check function calls, standardize arguments,
and return standardized call.
"""
import ast
import inspect
import builtins
from typing import Union
# add any libraries you might need for grading
# TODO we need a better way to include lib dependencies for running student code
from math import *

from .message_generators import (
    missing_argument, unexpected_argument, surplus_argument
)
from .formatters import formatted

def standardize_arguments(
    left_call: ast.Call, 
    right_call: ast.Call,
    left_source: str = "",
    right_source: str = "") -> Union[ast.Call, None]:
    """This will standardize the function calls for a function Call and return
    the modified Call.
    
    Returns None if there are no issues with the function call.
    
    Otherwise, an error message is returned if the user code can't be parsed,
    if there is a problem not finding variables in environment, or if arguments
    cannot be standardized.

    Parameters
    ----------
    left_call : ast.Call
        the user's function call to standardize arguments
    right_call : ast.Call
        the solution function call to standardize arguments
    left_source : str, optional
        the entire source code in which left_call belongs, by default ""
    right_source : str, optional
        the entire source code in which right_call belongs the source code, by default ""

    Returns
    -------
    ast.Call
        standardized function call

    Raises
    ------
    AssertionError
        when there is an issue with the standardization process
    """
    final_call = None
    try:
        # 1) introduce function call in environment
        # first pass: is it legal Python code?
        # for e.g., positional args should always be before keywords args
        # exec will catch these issues
        # NOTE: we are attempting to import `r` object for learnr use.
        if "r" in globals():
            exec(left_source, {}, r)
        else:
            exec(left_source)
        # 2) construct an environment containing both global and locals, 
        # overwriting the globals with locals, and builtins.
        envir = dict(globals(), **locals(), **builtins.__dict__)
        # 3) grab the live function from the environment
        # if we're calling a function on an object, the function info has
        # to be extracted from an ast.Attribute
        if isinstance(left_call.func, ast.Attribute):
            func_name = getattr(left_call.func, "attr")
            # the object's name 
            obj = left_call.func.value.id
            # the live function from the environment associated with object
            # e.g. the `df` in df.head() is a pd.DataFrame so must get the
            # `head` associated with that class.
            live_func = getattr(envir[obj], func_name)
        else:
            # if we're calling a function not associated with an object
            # just grab function from environment using its name
            func_name = left_call.func.id
            live_func = envir[func_name]

        # 4) collect the arguments passed
        # construct keyword args mapping
        kwargs = {a.arg: a.value for a in left_call.keywords}

        # 5) get the formal arguments for function
        sig = None
        try:
            # Note: this will raise ValueError if inspect cannot retrieve signature
            # which can happen for some builtins like `print`, where underlying C
            # code does not provide any metadata about its signature.
            sig = inspect.signature(live_func)
        except ValueError: 
            # if we can't get a signature just return call for normal Call
            # checking flow
            final_call = left_call
        if sig is None:
            raise ValueError()

        # 6) unpack args and kwargs and attempt to standardize argument calls
        # returns: https://docs.python.org/3.6/library/inspect.html#inspect.BoundArguments
        partial_args = sig.bind(*left_call.args, **kwargs)
        partial_args.apply_defaults()

        # 7) return a modified AST representing the standardized call
        # we do this by updating Call.keywords and reset the args
        new_keywords = []
        for k, v in partial_args.arguments.items():
            new_keywords.append(ast.keyword(arg=k, value=v))
        left_call.args = []
        left_call.keywords = new_keywords
        final_call = left_call
    except TypeError as e:
        # when call itself is invalid (e.g. wrong argument name)
        error = str(e)
        if "missing" in error:
            missing_argument(
                left_call,
                formatted(standardize_arguments(
                    left_call=right_call, right_call = right_call, 
                    left_source=right_source, right_source=right_source
                    )
                ),
                error
            )
        if "unexpected" in error:
            unexpected_argument(
                left_call,
                formatted(standardize_arguments(
                    left_call=right_call, right_call = right_call, 
                    left_source=right_source, right_source=right_source
                    )
                ),
                error
            )
        if "too many" in error:
            surplus_argument(
                left_call,
                formatted(standardize_arguments(
                    left_call=right_call, right_call = right_call, 
                    left_source=right_source, right_source=right_source
                    )
                ),
                error
            )
    except NameError as e:
        # when the function name doesn't exist in environment
        raise AssertionError(str(e).capitalize())
    except Exception as e:
        # when other Exceptions occur, silently ignore for now
        pass
    return final_call



"""
This module holds functions for common mistakes and generating messages for
feedback to the student.
"""

import ast
from typing import Dict, Any
from .formatters import formatted

def missing(left: Any, right: Any, line_info: Dict[str, int]) -> None:
    """Generates message when user is missing a node or element in code.

    Parameters
    ----------
    left : Any
        an ast.Node, a custom type, or a python data type
    right : Any
        an ast.Node, a custom type, or a python data type
    line_info : Dict[str, int]
        holds line information about a particular node

    Returns
    -------
    str
        feedback message
    """
    msg = "I expected `{}` at line {}."
    msg_args = (
        formatted(right), 
        line_info.get("left")
    )
    assert type(left) == type(right), msg.format(*msg_args)

def not_expected(left: Any, right: Any, line_info: Dict[str, int]) -> None:
    """Generates message when user supplies an extra node or element in code.

    Parameters
    ----------
    left : Any
        an ast.Node, a custom type, or a python data type
    right : Any
        an ast.Node, a custom type, or a python data type
    line_info : Dict[str, int]
        holds line information about a particular node

    Returns
    -------
    str
        feedback message
    """
    msg = "I did not expect `{}` at line {}."
    msg_args = (
        formatted(left), 
        line_info.get("left")
    )
    assert type(left) == type(right), msg.format(*msg_args)

def wrong_value(left: Any, right: Any, line_info: Dict[str, int], condition: bool, last_parent: str) -> None:
    """Generates message when user's code contains in an incorrect value.

    Parameters
    ----------
    left : Any
        an ast.Node, a custom type, or a python data type
    right : Any
        an ast.Node, a custom type, or a python data type
    line_info : Dict[str, int]
        holds line information about a particular node
    last_parent : str
        the nearest parent that can be converted to source text
    condition : bool
        the condition we want to check regarding equality between left and right
        ast nodes or values
    
    Returns
    -------
    str
        feedback message
    """
    msg = "I expected `{}`, but what you wrote was interpreted as `{}` at line {}."
    msg_args = (
        formatted(right),
        formatted(left),
        line_info.get("left"),
    )
    if last_parent != "" and last_parent != formatted(left):
        msg = "I expected `{}`, but what you wrote was interpreted as `{}` in `{}` at line {}."
        msg_args = (
            formatted(right),
            formatted(left),
            last_parent,
            line_info.get("left")
        )
    assert condition, msg.format(*msg_args)

# Call related -------------------------------

def repeated_argument(e: SyntaxError) -> str:
    """Generates a feedback for a repeated keyword argument error.

    Parameters
    ----------
    e : SyntaxError
        the error object

    Returns
    -------
    str
        feedback message
    """
    msg = "I couldn't parse your function call because I got a repeated keyword argument."
    return msg.format(str(e).lower())

def missing_argument(left_call: ast.AST, right_source: str, e: str) -> None:
    """Generates a feedback when there is a missing argument for a function call.

    Parameters
    ----------
    left_call : ast.AST
        student's function call
    right_source : str
        solution source text
    e : TypeError
        the error object

    Raises
    ------
    AssertionError
        to ripple up back to grade_code
    """
    msg = "I expected `{}` but what you wrote was interpreted as `{}`, which I can't execute because I'm {}."
    msg_args = (
        right_source, 
        formatted(left_call), 
        str(e).lower()
    )
    raise AssertionError(msg.format(*msg_args))

def unexpected_argument(left_call: ast.AST, right_source: str, e: str) -> None:
    """Generates a feedback when there is an unexpected argument for a function call.

    Parameters
    ----------
    left_call : ast.AST
        student's function call
    right_source : str
        solution source text
    e : TypeError
        the error object
    
    Raises
    ------
    AssertionError
        to ripple up back to grade_code
    """
    msg = "I expected `{}` but what you wrote was interpreted as `{}`, which I can't execute because I {}."
    msg_args = (
        right_source, 
        formatted(left_call), 
        str(e).lower()
    )
    raise AssertionError(msg.format(*msg_args))

def surplus_argument(left_call: ast.AST, right_source: str, e: str) -> None:
    """Generates a feedback when there extra arguments for a function call.

    Parameters
    ----------
    left_call : ast.AST
        student's function call
    right_source : str
        solution source text
    e : TypeError
        the error object
        
    Raises
    ------
    AssertionError
        to ripple up back to grade_code
    """
    msg = "I expected `{}` but what you wrote was interpreted as `{}`, which I can't execute because there are `{}`."
    msg_args = (
        right_source, 
        formatted(left_call), 
        str(e).lower()
    )
    raise AssertionError(msg.format(*msg_args))



from copy import deepcopy
from .ast_to_xml import xml
from .grade_code_found import GradeCodeFound
from .find_utils import uses, flatten_list

def find_functions(code: str | GradeCodeFound, match: str = "") -> GradeCodeFound:
  """Find function calls in the code.

  Parameters
  ----------
  code : str | GradeCodeFound
      a string for the source code, or a GradeCodeFound for chaining queries
  match : str, optional
      a particular function name, by default None

  Returns
  -------
  list[Element]
      list of XML elements corresponding to function definitions
  
  Examples
  --------
  >>> code = 'sum([1,2])\nsum([1,2,3])\nlen([1,2,3])'
  >>> find_functions(code)
  ── pygradecode found ──
  sum([1,2])
  sum([1,2,3])
  len([1,2,3])

  ── Result 1 ──
  sum([1,2])

  ── Result 2 ──
  sum([1,2,3])

  ── Result 3 ──
  len([1,2,3])
  >>> find_functions(code, "sum")
  ── pygradecode found ──
  sum([1,2])
  sum([1,2,3])
  len([1,2,3])

  ── Result 1 ──
  sum([1,2])

  ── Result 2 ──
  sum([1,2,3])
  """
  if not isinstance(code, str) and not isinstance(code, GradeCodeFound):
    raise Exception("`code` should be a `str` or `GradeCodeFound`")
  
  gcf = deepcopy(code) if isinstance(code, GradeCodeFound) else GradeCodeFound(code)
  xml_tree = xml(gcf.source) if isinstance(code, GradeCodeFound) else xml(code)

  request_type = 'functions'
  request = match
  result = []

  # TODO support a list of matches
  if request != "":
    xpath_query = f'.//Call//func/Name/id[.="{request}"]|.//attr[text()="{request}"]/ancestor::Call'
    if gcf.has_previous_request():
      for node in gcf.last_result:  
        id_nodes = node.xpath(xpath_query)
        if len(id_nodes) > 0:
          # grab the parent of the id element in order to view the source text
          # since id is not an ast.AST
          result.extend(flatten_list([get_ancestor_call(n) for n in id_nodes]))
    else:
      id_nodes = xml_tree.xpath(xpath_query)
      if len(id_nodes) > 0:
        # grab the parent of the id element in order to view the source text
        # since id is not an ast.AST
        result  = [get_ancestor_call(n) for n in id_nodes]
  else:
    xpath_query = f'.//Call|.//attr/ancestor::Call'
    if gcf.has_previous_request():
      for node in gcf.last_result:  
        result.extend(node.xpath(xpath_query))
    else:
      result = xml_tree.xpath(xpath_query)

  return gcf.push(request_type=request_type, request=request, result=result)

def uses_function(code: str, match: str = "") -> bool:
  """Check if the code uses functions.

  Parameters
  ----------
  code : str
      the source code
  match : str, optional
      function name(s), by default None

  Returns
  -------
  bool
      True if found, False otherwise

  Examples
  --------
  >>> code = 'sum([1,2,3])'
  >>> uses_function(code, "sum")
  True
  >>> uses_function(code, "round")
  False
  """
  return uses(find_functions, code, match)

def find_lambdas(code: str | GradeCodeFound) -> GradeCodeFound:
  """Check if there are lambdas in the code.

  Parameters
  ----------
  code : str | GradeCodeFound
      a string for the source code, or a GradeCodeFound for chaining queries

  Returns
  -------
  GradeCodeFound
      a GradeCoundFound object that holds the list of previous results
      and the current query results if any

  Examples
  --------
  >>> code = r'print(1, "2", sep=", ", end=foo(x = ["\n"]))'
  >>> find_lambdas(code)
  ── pygradecode found ──

  add_two = lambda x: x + 2
  add_two(1)


  Found 1 result

  ── Result 1 ──
  add_two = lambda x: x + 2
  """
  if not isinstance(code, str) and not isinstance(code, GradeCodeFound):
    raise Exception("`code` should be a `str` or `GradeCodeFound`")
  
  gcf = code if isinstance(code, GradeCodeFound) else GradeCodeFound(code)
  xml_tree = xml(gcf.source) if isinstance(code, GradeCodeFound) else xml(code)

  request_type = 'lambdas'
  result = []

  if gcf.has_previous_request():
    result = flatten_list([r.xpath(".//Lambda") for r in gcf.last_result])
  else:
    result = xml_tree.xpath("//Lambda")

  return gcf.push(request_type=request_type, request='', result=result)

def uses_lambda(code: str) -> bool:
  """Find lambdas within code.

  Parameters
  ----------
  code : str
      the source code

  Returns
  -------
  bool
      True if found, False otherwise

  Examples
  --------
  >>> code = 'add_two = lambda x: x + 2'
  >>> uses_lambda(code)
  True
  >>> code = 'def add_two():\n  return x + 2'
  >>> uses_lambda(code)
  False
  """
  return uses(find_lambdas, code)

# helper function to get the function <Call> given the <id> of function
def get_ancestor_call(node):
  # return current node if it is a Call
  if node.tag == 'Call':
    return node
  
  # recurse on parent
  return get_ancestor_call(node.getparent())



import ast
import textwrap
from typing import Optional, Dict, Any

from lxml import etree as ET
from lxml.etree import _Element as Element
from lxml.etree import SubElement

ATTRS = ("lineno", "col_offset", "end_lineno", "end_col_offset")

def ast_node_attrs(ast_node: ast.AST) -> Dict[str, Any]:
  """Given an AST node, return all of its attributes

  Parameters
  ----------
  ast_node : ast.AST
      The AST node

  Returns
  -------
  Dict
      A dictionary for the attribute name : value
  """
  attrs = {}
  for key in ATTRS:
    try:
      attrs[key] = str(getattr(ast_node, key))
    except AttributeError:
      continue
  return attrs

def visit_node(
  ast_node: ast.AST,
  parent_xml_node: Optional[Element] = None,
  parser: ET.XMLParser = ET.XMLParser(remove_blank_text=True)
) -> Element:
  """Given an AST node, traverse the tree while creating an XML Element tree.

  Parameters
  ----------
  ast_node : ast.AST
      The root node
  parent_xml_node : Optional[Element], optional
      The parent node, by default None

  Returns
  -------
  Element
      The root Element
  """
  xml_node_name = ast_node.__class__.__name__

  if parent_xml_node is None:
    xml_node = parser.makeelement(xml_node_name)
  else:
    xml_node = SubElement(parent_xml_node, xml_node_name)

  xml_node.attrib.update(ast_node_attrs(ast_node))

  for key, value in ast_node.__dict__.items():
    # skip dunders for now
    if key.startswith("_") or key in ATTRS:
      continue
    if isinstance(value, ast.AST):
      sub_node = SubElement(xml_node, key)
      visit_node(value, sub_node)
    elif isinstance(value, list):
      # sometimes nodes are in a list (e.g. Module body is a [])
      # so here we only create subelements if they are all AST nodes
      if all(isinstance(x, ast.AST) for x in value):
        sub_node = SubElement(xml_node, key)
        for node in value:
          visit_node(node, sub_node)
    else:
      # for non-AST nodes, just set `type` and `text` value (content of tag)
      node = SubElement(xml_node, key)
      node.attrib["type"] = type(value).__name__
      node.text = str(value)

  return xml_node

def xml(code: str) -> Element:
  """Return an XML tree of Python source code representing the AST.

  Parameters
  ----------
  src : str
      source code text

  Returns
  -------
  list[Element]
      a list of Element(s)
  """
  ast_tree = ast.parse(code)
  return visit_node(ast_tree)


# Get source text ----

def get_source_lines(
  src_lines: list[str],
  node: Element,
  dedent: bool = True
) -> str:
  code = "\n".join(src_lines)

  # attempt to extract source code lines relevant to the target node
  try:
    # extract ranges
    start_lineno = int(node.attrib["lineno"]) - 1
    end_lineno = int(node.attrib["end_lineno"])
    
    # turn the relevant lines it into one string
    code = "\n".join(src_lines[start_lineno:end_lineno])
    if dedent:
      code = textwrap.dedent(code)
  except (AttributeError, KeyError, ValueError):
    # if unsuccessful, we will return the entire code itself
    pass

  return code



import itertools
from typing import Optional

from lxml.etree import _Element as Element

def get_ancestor_node(node: Optional[Element]) -> Optional[Element]:
  if node is None or (hasattr(node, "attrib") and len(node.attrib) > 0):
    return node

  return get_ancestor_node(node.getparent())

# general `uses()` function to check if `find_*()` finds results based on queries
def uses(function, *args, **kwargs):
  return len(function(*args, **kwargs).last_result) > 0

def flatten_list(alist: list | list[list]) -> list:
  return list(itertools.chain(*alist))